/*
*/

// standard utilities and systems includes
#include <oclUtils.h>
#include <shrQATest.h>
#include <iostream>
#include <vector>
#include <string>

#ifndef _WIN32
typedef uint64_t memsize_t;
#else
typedef unsigned __int64 memsize_t;
#endif

// Name of the file with the source code for the computation kernel
// *********************************************************************
const char* cSourceFile = "oclSimple3d.cl";

// Host buffers for demo
// *********************************************************************
float *I, *O;                   // Host buffers for I and O
float* Golden;                  // Host buffer for host golden processing cross check

// OpenCL Vars
cl_platform_id cpPlatform;      // OpenCL platform
cl_context cxGPUContext;        // OpenCL context
cl_command_queue cqCommandQueue;// OpenCL command queue
cl_uint uiNumDevices;           // OpenCL device count
cl_device_id* cdDevices;        // OpenCL device list    
cl_uint targetDevice = 0;       // Default Device to compute on
cl_uint uiNumDevsUsed = 1;      // Number of devices used in this sample      
cl_program cpProgram;           // OpenCL program
cl_kernel ckKernel;             // OpenCL kernel
cl_event ceEvent;               // OpenCL event
cl_mem cmIN, cmOUT;                // OpenCL buffers for M and W
size_t szGlobalWorkSize;        // Total # of work items in the 1D range
size_t szLocalWorkSize;         // # of work items in the 1D work group    
size_t szParmDataBytes;         // Byte size of context information
size_t szKernelLength;          // Byte size of kernel code
cl_int ciErrNum;                // Error code var
char* cPathAndName = NULL;      // var for full paths to data, src, etc.
char* cSourceCL = NULL;         // Buffer to hold source for compilation 
const char* cExecutableName = NULL;

// demo config vars
int width  = 256;               // Matrix width
int height = 256;               // Matrix height
int depth  = 256;               // Matrix depth

shrBOOL bNoPrompt = shrFALSE;
bool bPassFlag = true;          // accumulator for test status

// Forward Declarations
// *********************************************************************
void DefineGolden(float* pfData, int iSize, const float defaultValue );
shrBOOL CompareData( const float* reference, const float* data, const unsigned int len, const float epsilon );
void Cleanup (int iExitCode);
void (*pCleanup)(int) = &Cleanup;

// Main function 
// *********************************************************************
int main(int argc, char** argv)
{
  const bool test3D = true; // 3D true, 2D false

  shrQAStart(argc, argv);
  // get command line arg for quick test, if provided
  bNoPrompt = shrCheckCmdLineFlag(argc, (const char **)argv, "noprompt");

  // start logs
  cExecutableName = argv[0];
  shrSetLogFileName ("oclSimple3d.txt");
  shrLog("%s Starting...\n\n", argv[0]); 

  // calculate matrix height given GPU memory
  shrLog("Determining matrix height from available GPU mem...\n");

  // Allocate and initialize host arrays
  shrLog("Allocate and Init Host Mem...\n\n");

  unsigned int size = 0;
  if(test3D)
    size = width * height * depth;
  else
    size = width * height;

  unsigned int mem_size_IN = size * sizeof(float);
  I = (float*)malloc(mem_size_IN);

  unsigned int mem_size_OUT = size * sizeof(float);
  O = (float*)malloc(mem_size_OUT);

  shrFillArray(I, size);

  Golden = (float*)malloc(mem_size_OUT);

  DefineGolden(Golden, size, 1);

  //Get the NVIDIA platform
  shrLog("Get the Platform ID...\n\n");
  ciErrNum = oclGetPlatformID(&cpPlatform);
  oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

  //Get all the devices
  shrLog("Get the Device info and select Device...\n");
  ciErrNum = clGetDeviceIDs(cpPlatform, CL_DEVICE_TYPE_GPU, 0, NULL, &uiNumDevices);
  oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);
  cdDevices = (cl_device_id *)malloc(uiNumDevices * sizeof(cl_device_id) );
  ciErrNum = clGetDeviceIDs(cpPlatform, CL_DEVICE_TYPE_GPU, uiNumDevices, cdDevices, NULL);
  oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

  // Set target device and Query number of compute units on targetDevice
  shrLog(" # of Devices Available = %u\n", uiNumDevices); 
  if(shrGetCmdLineArgumentu(argc, (const char **)argv, "device", &targetDevice)== shrTRUE) 
  {
    targetDevice = CLAMP(targetDevice, 0, (uiNumDevices - 1));
  }
  shrLog(" Using Device %u: ", targetDevice); 
  oclPrintDevName(LOGBOTH, cdDevices[targetDevice]);  
  cl_uint num_compute_units;
  clGetDeviceInfo(cdDevices[targetDevice], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(num_compute_units), &num_compute_units, NULL);
  shrLog("\n # of Compute Units = %u\n\n", num_compute_units); 

  //Create the context
  shrLog("clCreateContext...\n"); 
  cxGPUContext = clCreateContext(0, uiNumDevsUsed, &cdDevices[targetDevice], NULL, NULL, &ciErrNum);
  oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

  // Create a command-queue
  shrLog("clCreateCommandQueue...\n"); 
  cqCommandQueue = clCreateCommandQueue(cxGPUContext, cdDevices[targetDevice], CL_QUEUE_PROFILING_ENABLE, &ciErrNum);
  oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

  // Allocate the OpenCL buffer memory objects for source and result on the device GMEM
  shrLog("clCreateBuffer (I) in device global memory, mem_size_m = %u)...\n", mem_size_IN); 
  cmIN = clCreateBuffer(cxGPUContext, CL_MEM_READ_ONLY, mem_size_IN, NULL, &ciErrNum);
  oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

  shrLog("clCreateBuffer (O) in device global memory, mem_size_m = %u)...\n", mem_size_OUT); 
  cmOUT = clCreateBuffer(cxGPUContext, CL_MEM_WRITE_ONLY, mem_size_OUT, NULL, &ciErrNum);
  oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

  // Read the OpenCL kernel in from source file
  shrLog("oclLoadProgSource (%s)...\n", cSourceFile); 
  cPathAndName = shrFindFilePath(cSourceFile, argv[0]);
  oclCheckErrorEX(cPathAndName != NULL, shrTRUE, pCleanup);
  cSourceCL = oclLoadProgSource(cPathAndName, "", &szKernelLength);
  oclCheckErrorEX(cSourceCL != NULL, shrTRUE, pCleanup);

  // Create the program
  shrLog("clCreateProgramWithSource...\n"); 
  cpProgram = clCreateProgramWithSource(cxGPUContext, 1, (const char **)&cSourceCL, &szKernelLength, &ciErrNum);

  // Build the program
  shrLog("clBuildProgram...\n"); 
  ciErrNum = clBuildProgram(cpProgram, uiNumDevsUsed, &cdDevices[targetDevice], NULL, NULL, NULL);
  //ciErrNum = clBuildProgram(cpProgram, uiNumDevsUsed, &cdDevices[targetDevice], "-cl-fast-relaxed-math", NULL, NULL);
  if (ciErrNum != CL_SUCCESS)
  {
    // write out standard error, Build Log and PTX, then cleanup and exit
    shrLogEx(LOGBOTH | ERRORMSG, ciErrNum, STDERROR);
    oclLogBuildInfo(cpProgram, oclGetFirstDev(cxGPUContext));
    oclLogPtx(cpProgram, oclGetFirstDev(cxGPUContext), "oclSimple3d.ptx");
    shrQAFinish(argc, (const char **)argv, QA_FAILED);
    Cleanup(EXIT_FAILURE); 
  }

  // --------------------------------------------------------
  // Core sequence... copy input data to GPU, compute, copy results back

  // Asynchronous write of data to GPU device
  shrLog("clEnqueueWriteBuffer (I)...\n\n"); 
  ciErrNum = clEnqueueWriteBuffer(cqCommandQueue, cmIN, CL_TRUE, 0, mem_size_IN, I, 0, NULL, NULL);
  oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

  // Kernels
  std::vector<std::string> kernels3D;
  kernels3D.push_back("simple3d_test1");
  kernels3D.push_back("simple3d_test2");
  kernels3D.push_back("simple3d_test3");
  kernels3D.push_back("simple3d_test4");
  kernels3D.push_back("simple3d_test5");

  std::vector<std::string> kernels2D;
  kernels2D.push_back("simple2d_test1");
  kernels2D.push_back("simple2d_test2");

  std::vector<std::string> kernels;

  if(test3D)
    kernels = kernels3D;
  else
    kernels = kernels2D;

  for (unsigned int k = 0; k < kernels.size(); k++)
  {
    shrLog("Running with Kernel %s...\n\n", kernels[k].c_str()); 

    // Clear result
    shrLog("  Clear result with clEnqueueWriteBuffer (W)...\n"); 
    memset(O, 0, mem_size_OUT);
    ciErrNum = clEnqueueWriteBuffer(cqCommandQueue, cmOUT, CL_TRUE, 0, mem_size_OUT, O, 0, NULL, NULL);
    oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

    // Create the kernel
    shrLog("  clCreateKernel...\n"); 
    if (ckKernel) {
      clReleaseKernel(ckKernel);
      ckKernel = 0;
    }
    ckKernel = clCreateKernel(cpProgram, kernels[k].c_str(), &ciErrNum);
    oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

    size_t gws3D[3], lws3D[3];
    size_t gws2D[2], lws2D[2];
    // 3D
    gws3D[0] = width;
    gws3D[1] = height;
    gws3D[2] = depth;
    lws3D[0] = 4;
    lws3D[1] = 4;
    lws3D[2] = 4;

    // 2D
    gws2D[0] = width;
    gws2D[1] = height;
    lws2D[0] = 8;
    lws2D[1] = 8;

    // Set and log Global and Local work size dimensions
    //szLocalWorkSize = 256;
    //if (k == 0)
    //  szGlobalWorkSize = shrRoundUp((int)szLocalWorkSize, height);  // rounded up to the nearest multiple of the LocalWorkSize
    //else
    //  // Some experiments should be done here for determining the best global work size for a given device
    //  // We will assume here that we can run 2 work-groups per compute unit
    //  szGlobalWorkSize = 2 * num_compute_units * szLocalWorkSize;
    //shrLog("  Global Work Size \t\t= %u\n  Local Work Size \t\t= %u\n  # of Work Groups \t\t= %u\n", 
    //  szGlobalWorkSize, szLocalWorkSize, (szGlobalWorkSize % szLocalWorkSize + szGlobalWorkSize/szLocalWorkSize)); 

    // Set the Argument values
    shrLog("  clSetKernelArg...\n\n");
    int n = 0;

    // M and arguments
    ciErrNum = clSetKernelArg(ckKernel,  n++, sizeof(cl_mem), (void*)&cmIN);

    unsigned int s13D[4];
    unsigned int s12D[2];
    s13D[0] = width;
    s13D[1] = height;
    s13D[2] = depth;
    s13D[3] = 1;

    s12D[0] = width;
    s12D[1] = height;

    if(test3D)
      ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint4), (void*)&s13D);
    else
      ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint2), (void*)&s12D);

    if(k == 1)
    { 
      if(test3D)
      {
        for(unsigned int i=0; i<2; i++)
        {
          float f[4];
          ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float4), (void*)&f);
        }
        for(unsigned int i=0; i<3; i++)
        {
          float a[16];
          ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float16), (void*)&a);
        }
      }
      else
      {
        for(unsigned int i=0; i<2; i++)
        {
          float f[2];
          ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float2), (void*)&f);
        }
        for(unsigned int i=0; i<4; i++)
        {
          float a[16];
          ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float16), (void*)&a);
        }
      }
    }
    else if(k == 2)
    {
      for(unsigned int i=0; i<2; i++)
      {
        unsigned int f[2];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint2), (void*)&f);
      }
      for(unsigned int i=0; i<3; i++)
      {
        unsigned int a[4];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint4), (void*)&a);
      }
      for(unsigned int i=0; i<2; i++)
      {
        unsigned int c[8];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint8), (void*)&c);
      }
    }
    else if(k == 3)
    {
      for(unsigned int i=0; i<2; i++)
      {
        float f[2];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float4), (void*)&f);
      }
      for(unsigned int i=0; i<6; i++)
      {
        float f[8];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float8), (void*)&f);
      }
    }
    else if(k == 4)
    {
      for(unsigned int i=0; i<10; i++)
      {
        float a[16];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float16), (void*)&a);
      }
    }

    // W and arguments
    ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_mem), (void*)&cmOUT);

    unsigned int s23D[4];
    unsigned int s22D[2];
    s23D[0] = width;
    s23D[1] = height;
    s23D[2] = depth;
    s23D[3] = 1;

    s22D[0] = width;
    s22D[1] = height;

    if(test3D)
      ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint4), (void*)&s23D);
    else
      ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint2), (void*)&s22D);

    if(k == 1)
    { 
      if(test3D)
      {
        for(unsigned int i=0; i<2; i++)
        {
          float f[4];
          ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float4), (void*)&f);
        }
        for(unsigned int i=0; i<3; i++)
        {
          float a[16];
          ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float16), (void*)&a);
        }
      }
      else
      {
        for(unsigned int i=0; i<2; i++)
        {
          float f[2];
          ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float2), (void*)&f);
        }
        for(unsigned int i=0; i<4; i++)
        {
          float a[16];
          ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float16), (void*)&a);
        }
      }
    }
    else if(k == 2)
    {
      for(unsigned int i=0; i<2; i++)
      {
        unsigned int f[2];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint2), (void*)&f);
      }
      for(unsigned int i=0; i<3; i++)
      {
        unsigned int a[4];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint4), (void*)&a);
      }
      for(unsigned int i=0; i<2; i++)
      {
        unsigned int c[8];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_uint8), (void*)&c);
      }
    }
    else if(k == 3)
    {
      for(unsigned int i=0; i<2; i++)
      {
        float f[2];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float4), (void*)&f);
      }
      for(unsigned int i=0; i<6; i++)
      {
        float f[8];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float8), (void*)&f);
      }
    }
    else if(k == 4)
    {
      for(unsigned int i=0; i<10; i++)
      {
        float a[16];
        ciErrNum |= clSetKernelArg(ckKernel, n++, sizeof(cl_float16), (void*)&a);
      }
    }

    oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

    // Launch 3d kernel
    shrLog("  clEnqueueNDRangeKernel (%s)...\n", kernels[k].c_str()); 
    
    if(test3D)
      ciErrNum = clEnqueueNDRangeKernel(cqCommandQueue, ckKernel, 3, NULL, gws3D, lws3D, 0, NULL, &ceEvent);
    else
      ciErrNum = clEnqueueNDRangeKernel(cqCommandQueue, ckKernel, 2, NULL, gws2D, lws2D, 0, NULL, &ceEvent);

    oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

    // Read back results and check accumulated errors
    shrLog("  clEnqueueReadBuffer (W)...\n"); 
    ciErrNum = clEnqueueReadBuffer(cqCommandQueue, cmOUT, CL_TRUE, 0, mem_size_OUT, O, 0, NULL, NULL);
    oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

#ifdef GPU_PROFILING
    // Execution time
    ciErrNum = clWaitForEvents(1, &ceEvent);
    oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);
    cl_ulong start, end;
    ciErrNum = clGetEventProfilingInfo(ceEvent, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
    ciErrNum |= clGetEventProfilingInfo(ceEvent, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
    oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);
    double dSeconds = 1.0e-9 * (double)(end - start);
    shrLog("  Kernel execution time: %.5f s\n\n", dSeconds);
#endif

    // Compare results for golden-host and report errors and pass/fail
    shrLog("  Comparing against Host/C++ computation...\n\n"); 
    shrBOOL res = CompareData(Golden, O, size, 1e-6f);
    shrLog("    GPU Result %s CPU Result within allowable tolerance\n\n", (res == shrTRUE) ? "MATCHES" : "DOESN'T MATCH");
    bPassFlag &= (res == shrTRUE); 

    // Release event
    ciErrNum = clReleaseEvent(ceEvent);
    oclCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);
    ceEvent = 0;
  }

  // Master status Pass/Fail (all tests)
  shrQAFinish(argc, (const char **)argv, (bPassFlag ? QA_PASSED : QA_FAILED) );

  // Cleanup and leave
  Cleanup (EXIT_SUCCESS);
}

// Cleanup and exit code
// *********************************************************************
void Cleanup(int iExitCode)
{
  // Cleanup allocated objects
  shrLog("Starting Cleanup...\n\n");
  if(cdDevices)free(cdDevices);
  if(cPathAndName)free(cPathAndName);
  if(cSourceCL)free(cSourceCL);
  if(ceEvent)clReleaseEvent(ceEvent);  
  if(ckKernel)clReleaseKernel(ckKernel);  
  if(cpProgram)clReleaseProgram(cpProgram);
  if(cqCommandQueue)clReleaseCommandQueue(cqCommandQueue);
  if(cxGPUContext)clReleaseContext(cxGPUContext);
  if (cmIN)clReleaseMemObject(cmIN);
  if (cmOUT)clReleaseMemObject(cmOUT);

  // Free host memory
  free(I); 
  free(O);
  free(Golden);

  // finalize logs and leave
  if (bNoPrompt) {
    shrLogEx(LOGBOTH | CLOSELOG, 0, "%s Exiting...\n", cExecutableName);
  } else {
    shrLogEx(LOGBOTH | CLOSELOG, 0, "%s Exiting...\nPress <Enter> to Quit\n", cExecutableName);
    getchar();
  }
  exit (iExitCode);
}

// "Golden" Host processing matrix vector multiplication function for comparison purposes
void DefineGolden( float* pfData, const int iSize, const float defaultValue )
{
  for(int i = 0; i < iSize; ++i) 
  {
    pfData[i] = defaultValue;
  }
}

shrBOOL CompareData( const float* reference, const float* data,
                    const unsigned int len, const float epsilon )
{
  ARGCHECK(epsilon >= 0);

  float error = 0;
  float ref = 0;
  unsigned int count = 0;

  for( unsigned int i = 0; i < len; ++i) {

    float diff = reference[i] - data[i];
    error += diff * diff;
    ref += reference[i] * reference[i];

    if(abs(reference[i] - data[i]) > epsilon)
      count++;
  }

  float normRef = sqrtf(ref);
  if (fabs(ref) < 1e-7) {
    std::cerr << "ERROR, reference l2-norm is 0\n";
    return shrFALSE;
  }
  float normError = sqrtf(error);
  error = normError / normRef;
  bool result = error < epsilon;
  if( ! result) 
  {
    std::cerr << "ERROR, l2-norm error " 
      << error << " is greater than epsilon " << epsilon << "\n";

    std::cerr << "ERROR, number of wrong pixels " 
      << count << " that is " << (((float)count / (float)len) * 100.0) << "% \n";
  }
  return result ? shrTRUE : shrFALSE;
}
